/*
    SPDX-FileCopyrightText: 2025 Kdenlive contributors
    SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
*/

#pragma once

namespace AudioLevel {

enum class LevelStyle { Solid, Gradient };

enum class PeakIndicatorStyle { Colorful, Monochrome };

enum class TickLabelsMode { Show, Hide, HideIfSpaceIsLimited };

} // namespace AudioLevel