/*
    SPDX-FileCopyrightText: 2012 Jean-Baptiste Mardelle <jb@kdenlive.org>

SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
*/

#include "abstractcollapsiblewidget.h"

AbstractCollapsibleWidget::AbstractCollapsibleWidget(QWidget *parent)
    : QWidget(parent)
{
    setupUi(this);
}
